@@ -9,6 +9,17 @@
 #                                #
 #  theft is treason, citizen     #
 #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<#
+# History:
+# $RCSfile: Base.pm,v $ $Revision: 1.3 $ $Date: 2002/05/31 13:18:02 $
+# $Author: dassing $
+# $Log: Base.pm,v $
+# Revision 1.3  2002/05/31 13:18:02  dassing
+# Release 1.1
+#
+# Revision 1.2  2002/05/29 16:13:20  dassing
+# Changes included by David Pottage
+#
+##################################
 
 package Chart::Base;
 
@@ -17,7 +28,7 @@ use strict;
 use Carp;
 use FileHandle;
 
-$Chart::Base::VERSION = 1.00;
+$Chart::Base::VERSION = '1.1';
 
 use vars qw(%named_colors);
 
@@ -50,6 +61,7 @@ sub set {
     carp "Whoops, some option to be set didn't have a value.\n",
          "You might want to look at that.\n";
   }
+ 
   
   # set the options
   for (keys %opts) {
@@ -434,8 +446,10 @@ sub _init {
   # use 20 pixel length example lines in the legend
   $self->{'legend_example_size'} = 20;
 
-  # use 6 ticks on the y-axis
+  # Set the maximum & minimum number of ticks to use.
   $self->{'y_ticks'} = 6;
+  $self->{'min_y_ticks'} = 6;
+  $self->{'max_y_ticks'} = 100;
 
   # make the ticks 4 pixels long
   $self->{'tick_len'} = 4;
@@ -472,6 +486,9 @@ sub _init {
 
   # don't force integer y-ticks
   $self->{'integer_ticks_only'} = 'false';
+  
+  # don't forbid a false zero scale.
+  $self->{'include_zero'} = 'false';
 
   # don't waste time/memory by storing imagemap info unless they ask
   $self->{'imagemap'} = 'false';
@@ -792,27 +809,6 @@ sub _draw {
 sub _set_colors {
   my $self = shift;
   
-## Moved out to module level.
-#   my %colors = ('white'	=> [255,255,255],
-#   		'black'		=> [0,0,0],
-# 		'red'		=> [200,0,0],
-# 		'green'		=> [0,175,0],
-# 		'blue'		=> [0,0,200],
-# 		'orange'	=> [250,125,0],
-# 		'yellow'	=> [225,225,0],
-# 		'purple'	=> [200,0,200],
-# 		'light_blue'	=> [0,125,250],
-# 		'light_green'	=> [125,250,0],
-# 		'light_purple'	=> [145,0,250],
-# 		'pink'		=> [250,0,125],
-# 		'peach'		=> [250,125,125],
-# 		'olive'		=> [125,125,0],
-# 		'plum'		=> [125,0,125],
-# 		'turquoise'	=> [0,125,125],
-# 		'mauve'		=> [200,125,125],
-# 		'brown'		=> [160,80,0],
-# 		'grey'		=> [225,225,225]);
-
 
   my $index = $self->_color_role_to_index('background'); # allocate GD color
   if ( $self->{'transparent'} =~ m/^true$/i ) {
@@ -898,20 +894,25 @@ sub _set_colors {
 #   return 1; 
 }
 
-  sub _color_role_to_index {
+sub _color_role_to_index {
     my $self = shift;
     
     # Return a (list of) color index(es) corresponding to the (list of) role(s) in @_.
     my @result =  map {
-      my $role = $_;
-      my $index = $self->{'color_table'}->{$role};
+    my $role = $_;
+    my $index = $self->{'color_table'}->{$role};
+    
+    #print STDERR "Role = $_\n"; 
       
       unless ( defined $index ) {
         my $spec = $self->{'colors'}->{$role} 
           || $self->{'colors_default_spec'}->{$role}
           || $self->{'colors_default_spec'}->{$self->{'colors_default_role'}->{$role}};
+          
+      
         my @rgb = $self->_color_spec_to_rgb($role, $spec);
-        
+        #print STDERR "spec = $spec\n";
+       
         my $string = sprintf " RGB(%d,%d,%d)", map { $_ + 0 } @rgb;
         
         $index = $self->{'color_table'}->{$string};
@@ -924,6 +925,7 @@ sub _set_colors {
       }
       $index;
     } @_;
+    #print STDERR "Result= ".$result[0]."\n";
     (wantarray && @_ > 1 ? @result : $result[0]);
   }
       
@@ -1014,6 +1016,9 @@ sub _draw_sub_title {
   # now draw the subtitle
   $self->{'gd_obj'}->string ($font, $x, $y, $text, $color);
 
+  # Adapt curr_y_min
+  $self->{'curr_y_min'} += $self->{'text_space'} + $h;
+
   # and return
   return 1;
 }
@@ -1024,154 +1029,257 @@ sub _sort_data {
 
 }
 
-
 ##  find good values for the minimum and maximum y-value on the chart
-sub _find_y_scale {
-  my $self = shift;
-  my $data = $self->{'dataref'};
-#   my ($i, $j, $max, $min);
-  my ($order, $mult, $tmp);
-  my ($range, $delta, @dec, $y_ticks);
-  my $labels = [];
-  my $length = 0;
-
-## Handle data lying outside of -999999999999 to 999999999999
-## Completely ignore missing data.
-## Simplify loops.
-## Put range finding in a separate method for overriding.
-#   # use realy improbable starting max and min values
-#   $max = -999999999999;
-#   $min = 999999999999;
+# New version, re-written by David Pottage of Tao Group.
+# This code is *AS IS* and comes with *NO WARRANTY*
+#
+# This Sub calculates correct values for the following class local variables,
+# if they have not been set by the user.
+#
+# max_val, min_val: 	The maximum and minimum values for the y axis.
 # 
-#   # get the real max and min values
-#   for $i (1..$#{$data}) {
-#     for $j (0..$#{$data->[$i]}) {
-#       if ($data->[$i][$j] > $max) {
-# 	$max = $data->[$i][$j];
-#       }
-#       if ($data->[$i][$j] < $min) {
-#         # skip undefined values, for these are 'no data' values
-#         if (defined($data->[$i][$j])) {
-#           $min = $data->[$i][$j];
-#         }
-#       }
-#     }
-#   }
+# y_ticks:				The number of ticks to plot on the y scale, including
+#						the end points. e.g. If the scale runs from 0 to 50,
+#						with ticks every 10, y_ticks will have the value of 6.
+# 
+# y_tick_labels:		An array of strings, each is a label for the y axis.
+# 
+# y_tick_labels_length:	The length to allow for B tick labels. (How long is
+#						the longest?)	
+sub _find_y_scale
+{
+	my $self = shift;
+	
+	# Predeclare vars.
+	my ($d_min, $d_max);		# Dataset min & max.
+	my $d_width;				# Dataset width.
+	my ($p_min, $p_max);		# Plot min & max.
+	my ($tickInterval, $tickCount);
+	my @tickLabels;				# List of labels for each tick.
+	my $maxtickLabelLen = 0;	# The length of the longest tick label.
+	
+	# Find the datatset minimum and maximum.
+	($d_min, $d_max) = $self->_find_y_range();
+	
+	
+	# Force the inclusion of zero if the user has requested it.
+	if( $self->{'include_zero'} =~ m!^true$!i )
+	{
+		if( ($d_min * $d_max) > 0 )	# If both are non zero and of the same sign.
+		{
+			if( $d_min > 0 )	# If the whole scale is positive.
+			{
+				$d_min = 0;
+			}
+			else				# The scale is entirely negative.
+			{
+				$d_max = 0;
+			}
+		}
+	}
+	
+	# Calculate the width of the dataset.
+	$d_width = $d_max - $d_min;
+	
+	# If the width of the range is zero, forcibly widen it
+	# (to avoid division by zero errors elsewhere in the code).
+	if( 0 == $d_width )
+	{
+		$d_min--;
+		$d_max++;
+		$d_width = 2;
+	}
 
-  my ($min, $max) = $self->_find_y_range();
+        if ( $self->{'integer_ticks_only'} =~ /^\d$/ ) {
+           if ( $self->{'integer_ticks_only'} == 1 ) {
+              $self->{'integer_ticks_only'} = 'true';
+           } else {
+              $self->{'integer_ticks_only'} = 'false';
+           }
+        }
+           	
+	if( $self->{'integer_ticks_only'} =~ m!^true$!i )
+	{
+		# The user asked for integer ticks, force the limits to integers.
+		# & work out the range directly.
+		$p_min = _round2Tick($d_min, 1, -1);
+		$p_max = _round2Tick($d_max, 1, 1);
+		
+		$tickInterval = 1;
+		$tickCount = $p_max - $p_min +1;
+	}
+	else
+	{
+                # Allow the dataset range to be overridden by the user.
+                # f_min/f_max are booleans which indicate the min & max sould not be modified.
+                my $f_min = defined $self->{'min_val'};
+                $d_min = $self->{'min_val'} if $f_min;
+                
+                my $f_max = defined $self->{'max_val'};
+                $d_max = $self->{'max_val'} if $f_max;
+                
+		# Descale the range by converting the dataset width into
+		# a floating point exponent & mantisa pair.
+		my( $rangeExponent, $rangeMantisa ) = _sepFP( $d_width );
+		my $rangeMuliplier = 10 ** $rangeExponent;
 
-  # calculate good max_val
-  if ($max < -10) {
-    $tmp = -$max;
-    $order = int((log $tmp) / (log 10));
-    $mult = int ($tmp / (10 ** $order));
-    $tmp = ($mult - 1) * (10 ** $order);
-    $max = -$tmp;
-  }
-  elsif ($max < 0) {
-    $max = 0;
-  }
-  elsif ($max > 10) {
-    $order = int((log $max) / (log 10));
-    $mult = int ($max / (10 ** $order));
-    $max = ($mult + 1) * (10 ** $order);
-  }
-  elsif ($max >= 0) {
-    $max = 10;
-  }
+		# Find what tick interval to use & how many ticks to plot,
+		# round the plot min & max to suitable round numbers.
+		($tickInterval, $tickCount, $p_min, $p_max)
+			= _calcTickInterval($d_min/$rangeMuliplier, $d_max/$rangeMuliplier,
+								$f_min, $f_max,
+								$self->{'min_y_ticks'}, $self->{'max_y_ticks'});
 
-  # now go for a good min_val
-  if ($min < -10) {
-    $tmp = -$min;
-    $order = int((log $tmp) / (log 10));
-    $mult = int ($tmp / (10 ** $order));
-    $tmp = ($mult + 1) * (10 ** $order);
-    $min = -$tmp;
-  }
-  elsif ($min < 0) {
-    $min = -10;
-  }
-  elsif ($min > 10) {
-    $order = int ((log $min) / (log 10));
-    $mult = int ($min / (10 ** $order));
-    $min = $mult * (10 ** $order);
-  }
-  elsif ($min >= 0) {
-    $min = 0;
-  }
+		# Restore the tickInterval etc to the correct scale
+		$_ *= $rangeMuliplier foreach($tickInterval, $p_min, $p_max);
+	}
+	
+	# Now sort out an array of tick labels.
+	for( my $labelNum = $p_min; $labelNum<=$p_max; $labelNum+=$tickInterval )
+	{
+		my $labelText;
+		
+		if( defined $self->{f_y_tick} )
+		{
+			$labelText = $self->{f_y_tick}->($labelNum);
+		}
+		else
+		{
+			$labelText = sprintf("%.3f", $labelNum);
+		}
+		
+		push @tickLabels, $labelText;
+		$maxtickLabelLen = length $labelText if $maxtickLabelLen < length $labelText;
+	}
+	
+	# Store the calculated data.
+	$self->{'min_val'} = $p_min;
+	$self->{'max_val'} = $p_max;
+	$self->{'y_ticks'} = $tickCount;
+	$self->{'y_tick_labels'} = \@tickLabels;
+	$self->{'y_tick_label_length'} = $maxtickLabelLen;
+	
+	# and return.
+	return 1;
+}
 
-  # put min_val and max_val into the object unless they're already there
-  unless (defined ($self->{'max_val'})) {
-    $self->{'max_val'} = $max;
-  }
-  unless (defined ($self->{'min_val'})) {
-    $self->{'min_val'} = $min;
-  }
+# Calculates the tick interval in normalised units.
+# Result will need multiplying by the multipler to get the true tick interval.
+# written by David Pottage of Tao Group.
+sub _calcTickInterval
+{
+	my(
+		$min, $max,		# The dataset min & max.
+		$minF, $maxF,	# Indicates if those min/max are fixed.
+		$minTicks, $maxTicks,	# The minimum & maximum number of ticks.
+	) = @_;
+	
+	# Verify the supplied 'min_y_ticks' & 'max_y_ticks' are sensible.
+	if( $minTicks < 2 )
+	{
+		print STDERR "Chart::Base : Incorrect value for 'min_y_ticks', too small.\n";
+		$minTicks = 2;
+	}
+	
+	if( $maxTicks < 5*$minTicks  )
+	{
+		print STDERR "Chart::Base : Incorrect value for 'max_y_ticks', to small.\n";
+		$maxTicks = 5*$minTicks;
+	}
+	
+	my $width = $max - $min;
+	my @divisorList;
+	
+	for( my $baseMul = 1; ; $baseMul *= 10 )
+	{
+		TRY: foreach my $tryMul (1, 2, 5)
+		{
+			# Calc a fresh, smaller tick interval.
+			my $divisor = $baseMul * $tryMul;
+			
+			# Count the number of ticks.
+			my ($tickCount, $pMin, $pMax) = _countTicks($min, $max, 1/$divisor);
+			
+			# Look a the number of ticks.
+			if( $maxTicks < $tickCount )
+			{
+				# If it is to high, Backtrack.
+				$divisor = pop @divisorList;
+				($tickCount, $pMin, $pMax) = _countTicks($min, $max, 1/$divisor);
+				print "\nChart::Base : Caution: Tick limit of $maxTicks exceeded. Backing of to an interval of ".1/$divisor." which plots $tickCount ticks\n";
+				return(1/$divisor, $tickCount, $pMin, $pMax);
+			}
+			elsif( $minTicks > $tickCount )
+			{
+				# If it is to low, try again.
+				next TRY;
+			}
+			else
+			{
+				# Store the divisor for possible later backtracking.
+				push @divisorList, $divisor;
+			
+				# if the min or max is fixed, check they will fit in the interval.
+				next TRY if( $minF && ( int ($min*$divisor) != ($min*$divisor) ) );
+				next TRY if( $maxF && ( int ($max*$divisor) != ($max*$divisor) ) );
+				
+				# If everything passes the tests, retrun.
+				return(1/$divisor, $tickCount, $pMin, $pMax)
+			}
+		}
+	}
+	
+	die "can't happen!";
+}
 
-  # find the range of the y-axis, modify the number of y-ticks
-  # if they asked for integer ticks only
-  $range = $self->{'max_val'} - $self->{'min_val'};
-  $y_ticks = $self->{'y_ticks'} - 1;
-  ## Don't adjust y_ticks if the user specified custom labels
-  if ($self->{'integer_ticks_only'} =~ /^true$/i && ! $self->{'y_tick_labels'}) {
-    unless (($range % $y_ticks) == 0) {
-      while (($range % $y_ticks) != 0) {
-	$y_ticks++;
-      }
-      $self->{'y_ticks'} = $y_ticks + 1;
-    }
-  }
+# Works out how many ticks would be displayed at that interval
+# e.g min=2, max=5, interval=1, result is 4 ticks.
+# written by David Pottage of Tao Group.
+sub _countTicks
+{
+	my( $min, $max, $interval) = @_;
+	
+	my $minR = _round2Tick( $min, $interval, -1);
+	my $maxR = _round2Tick( $max, $interval, 1);
+	
+	my $tickCount = ( $maxR/$interval ) - ( $minR/$interval ) +1;
+	
+	return ($tickCount, $minR, $maxR);
+}
 
-## Allow custom y_tick_labels, see below.
-#   # generate the y-tick labels, find the longest one
-#   $delta = $range / $y_ticks;
-#   for (0..$y_ticks) {
-#     $tmp = $self->{'min_val'} + ($delta * $_);
-#     @dec = split /\./, $tmp;
-#     if ($dec[1] && (length($dec[1]) > 3)) {
-#       $tmp = sprintf("%.3f", $tmp);
-#     }
-#     $labels->[$_] = $tmp;
-#     if (length($tmp) > $length) {
-#       $length = length($tmp);
-#     }
-#   }
+# Rounds up or down to the next tick of interval size.
+# $roundUP can be +1 or -1 to indicate if rounding should be up or down.
+# written by David Pottage of Tao Group.
+sub _round2Tick
+{
+	my($input, $interval, $roundUP) = @_;
+	return $input if $interval == 0;
+	die unless 1 == $roundUP*$roundUP;
+	
+	my $intN  = int ($input/$interval);
+	my $fracN = ($input/$interval) - $intN;
+	
+	my $retN = ( ( 0 == $fracN ) || ( ($roundUP * $fracN) < 0 ) )
+		? $intN
+		: $intN + $roundUP;
+	
+	return $retN * $interval;
+}
 
-  $labels = $self->{'y_tick_labels'};
-  unless ( $labels ) { 
-    # generate the y-tick labels, find the longest one
-    $delta = $range / $y_ticks;
-    for (0..$y_ticks) {
-      $tmp = $self->{'min_val'} + ($delta * $_);
-      @dec = split /\./, $tmp;
-      if ($dec[1] && (length($dec[1]) > 3)) {
-        $tmp = sprintf("%.3f", $tmp);
-      }
-      $labels->[$_] = $tmp;
-    }
-  }
-  else { # custom y tick labels
-    croak "Need $self->{'y_ticks'} y_tick_labels, got " . scalar(@$labels) . "\n"
-      unless @$labels == $self->{'y_ticks'};
-   
-  }
-#  for my $label ( @$labels ) {
-#    if (length($label) > $length) {
-#      $length = length($label);
-#    }
-#  }
-  for my $label ( @$labels ) {
-    if (length($self->{f_y_tick}->($label)) > $length) {
-      $length = length($self->{f_y_tick}->($label));
-    }
-  }
-
-  # store it in the object
-  $self->{'y_tick_labels'} = $labels;
-  $self->{'y_tick_label_length'} = $length;
- 
-  # and return
-  return 1;
+# Seperates a number into it's base 10 floating point exponent & mantisa.
+# written by David Pottage of Tao Group.
+sub _sepFP
+{
+	my($num) = @_;
+	return(0,0) if $num == 0;
+	
+	my $sign = ( $num > 0 ) ? 1 : -1;
+	$num *= $sign;
+	
+	my $exponent = int ( log($num)/log(10) );
+	my $mantisa  = $sign *($num / (10**$exponent) );
+	
+	return ( $exponent, $mantisa );
 }
 
 sub _find_y_range {
@@ -1279,7 +1387,7 @@ sub _draw_legend {
   elsif ($self->{'legend'} eq 'top') {
     $self->_draw_top_legend;
   } else {
-    carp "I can't put a legend there\n";
+    carp "I can't put a legend there (at ".$self->{'legend'}.")\n";
   }
 
   # and return
@@ -1366,15 +1474,15 @@ sub _draw_bottom_legend {
 
         # reset the brush for points
         $brush = $self->_prepare_brush($color, 'point',
-				$self->{'pointStyle' . $_});
+				$self->{'pointStyle' . $index});
         $self->{'gd_obj'}->setBrush($brush);
         # draw the point
-        $self->{'gd_obj'}->line(int(($x3+$x2)/2), $y2,
-	  	int(($x3+$x2)/2), $y2, gdBrushed);
+        $x3 = int($x + $self->{'legend_example_size'}/2);
+        $self->{'gd_obj'}->line($x3, $y, $x3, $y, gdBrushed);
 
         # adjust the x-y coordinates for the start of the label
 	$x += $self->{'legend_example_size'} + (2 * $self->{'text_space'});
-	$y -= $h/2;
+        $y = $y1 + ($row_height * $r);
 
 	# now draw the label
 	$self->{'gd_obj'}->string($font, $x, $y, $labels[$index], $color);
@@ -1546,11 +1654,11 @@ sub _draw_top_legend {
 
         # reset the brush for points
         $brush = $self->_prepare_brush($color, 'point',
-				$self->{'pointStyle' . $_});
+				$self->{'pointStyle' . $index});
         $self->{'gd_obj'}->setBrush($brush);
         # draw the point
-        $self->{'gd_obj'}->line(int(($x3+$x2)/2), $y2,
-	  	int(($x3+$x2)/2), $y2, gdBrushed);
+        $x3 = int($x + $self->{'legend_example_size'}/2);
+        $self->{'gd_obj'}->line($x3, $y, $x3, $y, gdBrushed);
 
         # now the label
 	$x += $self->{'legend_example_size'} + (2 * $self->{'text_space'});
